//
//  fbc.cpp
//  Algorithm: FBC128_256
//
//  Created by administrator on 2019/10/15.
//  Copyright © 2019 X. Feng. All rights reserved.
//

#include <assert.h>
#include "fbc.h"

//#include <stdio.h>
//void print2(unsigned char* chs, int len) {
//    for (int i = 0; i < len; i ++) {
//        printf("%02x ", chs[i]);
//    }
//    printf("\n");
//}


typedef unsigned int word;

const int FBC128_256_BLOCK_LEN = 16;
const int FBC128_256_KEY_LEN = 32;
const int FBC128_256_R = 64;
const int FBC128_256_SUBKEY_LEN = 2 * FBC128_256_R;
const word full_one_mask = -1;

#define L(x, s, t)    ((x) ^ (((x) << s) | ((x) >> (32 - s))) ^ (((x) << t) | ((x) >> (32 - t))))

#define F32(x)          u[0] = x & 0xFF; u[1] = (x >> 8) & 0xFF; u[2] = (x >> 16) & 0xFF; u[3] = (x >> 24) & 0xFF;     \
                        v[0] = (u[0] ^ 0xFF); u[0] = u[2] ^ 0xFF; v[0] ^= u[3] & u[0];      \
                        v[1] = (u[1] ^ 0xFF); u[1] = u[3] ^ 0xFF; v[1] ^= v[0] & u[1];      \
                        v[2] = u[0] ^ (v[1] & (v[0] ^ 0xFF));                               \
                        v[3] = u[1] ^ (v[2] & (v[1] ^ 0xFF));                               \
                        x = v[0] | (v[1] << 8) | (v[2] << 16) | (v[3] << 24);               \
                        x = L(x, 3, 10);

int Key_Schedule(unsigned char* seedkey, int keylen, int direction, unsigned char* subkey) {
    assert(keylen == FBC128_256_KEY_LEN);

    word* sk = (word *) &seedkey[0];
    word* rk = (word *) &subkey[0];
    int i;

    rk[0] = sk[0]; rk[1] = sk[1]; rk[2] = sk[2]; rk[3] = sk[3];
    rk[4] = sk[4]; rk[5] = sk[5]; rk[6] = sk[6]; rk[7] = sk[7];
    for (i = 0; i < FBC128_256_SUBKEY_LEN - 8; i ++) {
         rk[i+8] = (rk[i] ^ full_one_mask) ^ (((rk[i+1] ^ full_one_mask) & rk[i+2]) & rk[i+3]) ^ i;
         rk[i+8] = L(rk[i+8], 13, 22);
    }

    return 0;
}

void Crypt_Enc_One_Block(unsigned char* input, unsigned char* output, word* rk) {
    word u[4], v[4], x, y;
    int  i = 0;

    word* a = (word *)&output[0];
    word* b = (word *)&input[0];
    a[0] = b[0]; a[1] = b[1]; a[2] = b[2]; a[3] = b[3];
    while (i < FBC128_256_SUBKEY_LEN - 3) {
        // the left F-function
        x = a[0] ^ rk[i ++];
        F32(x)
        x ^= a[1];

        // the right F-function
        y = a[3] ^ rk[i ++];
        F32(y)
        y ^= a[2];

        a[1] = a[0] ^ y;
        a[2] = a[3] ^ x;
        a[0] = x;
        a[3] = y;

//        print2(output, FBC128_256_BLOCK_LEN);
    }

    // the final round
    x = a[0] ^ rk[i ++];
    F32(x)
    a[1] ^= x;

    x = a[3] ^ rk[i];
    F32(x)
    a[2] ^= x;

    a[0] ^= a[2];
    a[3] ^= a[1];
}

void Crypt_Dec_One_Block(unsigned char* input, unsigned char* output, word* rk) {
    word u[4], v[4], x, y, c, d;
    int  i = FBC128_256_SUBKEY_LEN-3;

    word* a = (word *)&output[0];
    word* b = (word *)&input[0];
    a[0] = b[0]; a[1] = b[1]; a[2] = b[2]; a[3] = b[3];

    // the final round
    a[0] ^= a[2];
    a[3] ^= a[1];

    x = a[0] ^ rk[FBC128_256_SUBKEY_LEN-2];
    F32(x)
    a[1] ^= x;

    x = a[3] ^ rk[FBC128_256_SUBKEY_LEN-1];
    F32(x)
    a[2] ^= x;

    while (i > 0) {
        c = a[1] ^ a[3];
        d = a[0] ^ a[2];

        y = d ^ rk[i --];
        F32(y)
        a[2] = y ^ a[3];

        x = c ^ rk[i --];
        F32(x)
        a[1] = x ^ a[0];

        a[0] = c;
        a[3] = d;
     }
}

int Crypt_Enc_Block(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 15) == 0) && (keylen == FBC128_256_KEY_LEN));

    word rk[FBC128_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 0, (unsigned char*) rk);

    int i = 0;
    while (i < in_len) {
        Crypt_Enc_One_Block(&input[i], &output[i], rk);
        i += FBC128_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Dec_Block(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 15) == 0) && (keylen == FBC128_256_KEY_LEN));

    word rk[FBC128_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 1, (unsigned char*) rk);

    int i = 0;
    while (i < in_len) {
        Crypt_Dec_One_Block(&input[i], &output[i], rk);
        i += FBC128_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Enc_Block_CBC(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 15) == 0) && (keylen == FBC128_256_KEY_LEN));

    word rk[FBC128_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 0, (unsigned char*) rk);

    Crypt_Enc_One_Block(input, output, rk);

    word iv[4];
    int i = FBC128_256_BLOCK_LEN;
    while (i < in_len) {
        iv[0] = *((word *)(input+i)) ^ (*(word *)(output+i-FBC128_256_BLOCK_LEN));
        iv[1] = *((word *)(input+i+  sizeof(word))) ^ (*(word *)(output+i-FBC128_256_BLOCK_LEN+  sizeof(word)));
        iv[2] = *((word *)(input+i+2*sizeof(word))) ^ (*(word *)(output+i-FBC128_256_BLOCK_LEN+2*sizeof(word)));
        iv[3] = *((word *)(input+i+3*sizeof(word))) ^ (*(word *)(output+i-FBC128_256_BLOCK_LEN+3*sizeof(word)));

        Crypt_Enc_One_Block((unsigned char*)iv, &output[i], rk);
        i += FBC128_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Dec_Block_CBC(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 15) == 0) && (keylen == FBC128_256_KEY_LEN));

    word rk[FBC128_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 1, (unsigned char*) rk);

    Crypt_Dec_One_Block(input, output, rk);

    word iv[4];
    int i = FBC128_256_BLOCK_LEN;
    while (i < in_len) {
        Crypt_Dec_One_Block(&input[i], (unsigned char*) iv, rk);

        *((word *)(output+i)) = iv[0] ^ (*(word *)(input+i-FBC128_256_BLOCK_LEN));
        *((word *)(output+i+  sizeof(word))) = iv[1] ^ (*(word *)(input+i-FBC128_256_BLOCK_LEN+  sizeof(word)));
        *((word *)(output+i+2*sizeof(word))) = iv[2] ^ (*(word *)(input+i-FBC128_256_BLOCK_LEN+2*sizeof(word)));
        *((word *)(output+i+3*sizeof(word))) = iv[3] ^ (*(word *)(input+i-FBC128_256_BLOCK_LEN+3*sizeof(word)));

        i += FBC128_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Enc_Block_Round(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen, int cryptround) {
    assert((cryptround >= 1) && (cryptround <= FBC128_256_R) && (in_len == FBC128_256_BLOCK_LEN) && (keylen == FBC128_256_KEY_LEN));

    word* sk = (word *) key;
    word  rk[FBC128_256_SUBKEY_LEN];
    int   rk_num = 2 * cryptround;
    int   i = 0;

    rk[0] = sk[0]; rk[1] = sk[1]; rk[2] = sk[2]; rk[3] = sk[3];
    rk[4] = sk[4]; rk[5] = sk[5]; rk[6] = sk[6]; rk[7] = sk[7];
    for (i = 0; i < rk_num - 8; i ++) {
        rk[i+8] = (rk[i] ^ full_one_mask) ^ (((rk[i+1] ^ full_one_mask) & rk[i+2]) & rk[i+3]) ^ i;
        rk[i+8] = L(rk[i+8], 13, 22);
    }

    word  u[4], v[4], x, y;
    word* a = (word *)&output[0];
    word* b = (word *)&input[0];
    a[0] = b[0]; a[1] = b[1]; a[2] = b[2]; a[3] = b[3];
    while (i < rk_num - 3) {
        // the left F-function
        x = a[0] ^ rk[i ++];
        F32(x)
        x ^= a[1];

        // the right F-function
        y = a[3] ^ rk[i ++];
        F32(y)
        y ^= a[2];

        a[1] = a[0] ^ y;
        a[2] = a[3] ^ x;
        a[0] = x;
        a[3] = y;
    }

    // the final round
    x = a[0] ^ rk[i ++];
    F32(x)
    a[1] ^= x;

    x = a[3] ^ rk[i];
    F32(x)
    a[2] ^= x;

    a[0] ^= a[2];
    a[3] ^= a[1];

    *out_len = in_len;
    return 0;
}
