//
//  fbc.cpp
//  Algorithm: FBC256_256
//
//  Created by administrator on 2019/10/15.
//  Copyright © 2019 X. Feng. All rights reserved.
//

#include <assert.h>
#include "fbc.h"

#ifdef _WIN_32  // windows platform
    typedef unsigned __int64 word;
#else           // macos or linux platform
    typedef unsigned long long word;
#endif

const int FBC256_256_BLOCK_LEN = 32;
const int FBC256_256_KEY_LEN = 32;
const int FBC256_256_R = 80;
const int FBC256_256_SUBKEY_LEN = 2 * FBC256_256_R;
const word full_one_mask = -1;

#define L(x, s, t)   ((x) ^ (((x) << s) | ((x) >> (64 - s))) ^ (((x) << t) | ((x) >> (64 - t))))

#define F64(x)          u[0] = x & 0xFFFF; u[1] = (x >> 16) & 0xFFFF; u[2] = (x >> 32) & 0xFFFF; u[3] = (x >> 48) & 0xFFFF; \
                        v[0] = (u[0] ^ 0xFFFF); u[0] = u[2] ^ 0xFFFF; v[0] ^= u[3] & u[0];  \
                        v[1] = (u[1] ^ 0xFFFF); u[1] = u[3] ^ 0xFFFF; v[1] ^= v[0] & u[1];  \
                        v[2] = u[0] ^ (v[1] & (v[0] ^ 0xFFFF)); \
                        v[3] = u[1] ^ (v[2] & (v[1] ^ 0xFFFF)); \
                        x = v[0] | (v[1] << 16) | (v[2] << 32) | (v[3] << 48);  \
                        x = L(x, 17, 58);

// only for test
#include <stdio.h>
void print2(unsigned char* chs, int len) {
    for (int i = 0; i < len; i ++) {
        printf("%02x ", chs[i]);
    }
    printf("\n");
}

int Key_Schedule(unsigned char* seedkey, int keylen, int direction, unsigned char* subkey) {
    assert(keylen == FBC256_256_KEY_LEN);

    word* sk = (word *) &seedkey[0];
    word* rk = (word *) &subkey[0];
    int i;

    rk[0] = sk[0]; rk[1] = sk[1]; rk[2] = sk[2]; rk[3] = sk[3];
    for (i = 0; i < FBC256_256_SUBKEY_LEN - 4; i ++) {
         rk[i+4] = (rk[i] ^ full_one_mask) ^ (((rk[i+1] ^ full_one_mask) & rk[i+2]) & rk[i+3]) ^ i;
         rk[i+4] = L(rk[i+4], 13, 22);
    }

    return 0;
}

void Crypt_Enc_One_Block(unsigned char* input, unsigned char* output, word* rk) {
    word u[4], v[4], x, y;
    int  i = 0;

    word* a = (word *)&output[0];
    word* b = (word *)&input[0];
    a[0] = b[0]; a[1] = b[1]; a[2] = b[2]; a[3] = b[3];
    while (i < FBC256_256_SUBKEY_LEN-3) {
        // the left F-function
        x = a[0] ^ rk[i ++];
        F64(x)
        x ^= a[1];

        // the right F-function
        y = a[3] ^ rk[i ++];
        F64(y)
        y ^= a[2];

        a[1] = a[0] ^ y;
        a[2] = a[3] ^ x;
        a[0] = x;
        a[3] = y;

//        print2(output, FBC256_256_BLOCK_LEN);
    }

    // the final round
    x = a[0] ^ rk[i ++];
    F64(x)
    a[1] ^= x;

    x = a[3] ^ rk[i];
    F64(x)
    a[2] ^= x;

    a[0] ^= a[2];
    a[3] ^= a[1];
}

void Crypt_Dec_One_Block(unsigned char* input, unsigned char* output, word* rk) {
    word u[4], v[4], x, y, c, d;
    int  i = FBC256_256_SUBKEY_LEN-3;

    word* a = (word *)&output[0];
    word* b = (word *)&input[0];
    a[0] = b[0]; a[1] = b[1]; a[2] = b[2]; a[3] = b[3];

    // the final round
    a[0] ^= a[2];
    a[3] ^= a[1];

    x = a[0] ^ rk[FBC256_256_SUBKEY_LEN-2];
    F64(x)
    a[1] ^= x;

    x = a[3] ^ rk[FBC256_256_SUBKEY_LEN-1];
    F64(x)
    a[2] ^= x;

    while (i > 0) {
        c = a[1] ^ a[3];
        d = a[0] ^ a[2];

        y = d ^ rk[i --];
        F64(y)
        a[2] = y ^ a[3];

        x = c ^ rk[i --];
        F64(x)
        a[1] = x ^ a[0];

        a[0] = c;
        a[3] = d;
    }
}

int Crypt_Enc_Block(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 31) == 0) && (keylen == FBC256_256_KEY_LEN));

    word rk[FBC256_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 0, (unsigned char*) rk);

    int i = 0;
    while (i < in_len) {
        Crypt_Enc_One_Block(&input[i], &output[i], rk);
        i += FBC256_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Dec_Block(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 31) == 0) && (keylen == FBC256_256_KEY_LEN));

    word rk[FBC256_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 1, (unsigned char*) rk);

    int i = 0;
    while (i < in_len) {
        Crypt_Dec_One_Block(&input[i], &output[i], rk);
        i += FBC256_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Enc_Block_CBC(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 31) == 0) && (keylen == FBC256_256_KEY_LEN));

    word rk[FBC256_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 0, (unsigned char*) rk);

    Crypt_Enc_One_Block(input, output, rk);

    word m[4], *in, *iv;
    int i = FBC256_256_BLOCK_LEN;
    while (i < in_len) {
        in = (word *) &input[i];
        iv = (word *) &output[i-FBC256_256_BLOCK_LEN];
        m[0] = in[0] ^ iv[0]; m[1] = in[1] ^ iv[1]; m[2] = in[2] ^ iv[2]; m[3] = in[3] ^ iv[3];

        Crypt_Enc_One_Block((unsigned char*)m, &output[i], rk);
        i += FBC256_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Dec_Block_CBC(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen) {
    assert((in_len > 0) && ((in_len & 31) == 0) && (keylen == FBC256_256_KEY_LEN));

    word rk[FBC256_256_SUBKEY_LEN];
    Key_Schedule(key, keylen, 1, (unsigned char*) rk);

    Crypt_Dec_One_Block(input, output, rk);

    word m[4], *iv, *out;
    int i = FBC256_256_BLOCK_LEN;
    while (i < in_len) {
        Crypt_Dec_One_Block(&input[i], (unsigned char*) m, rk);

        out = (word *)&output[i];
        iv  = (word *)&input[i-FBC256_256_BLOCK_LEN];
        out[0] = m[0] ^ iv[0]; out[1] = m[1] ^ iv[1]; out[2] = m[2] ^ iv[2]; out[3] = m[3] ^ iv[3];

        i += FBC256_256_BLOCK_LEN;
    }

    *out_len = in_len;
    return 0;
}

int Crypt_Enc_Block_Round(unsigned char* input, int in_len, unsigned char* output, int* out_len, unsigned char* key, int keylen, int cryptround) {
    assert((cryptround >= 1) && (cryptround <= FBC256_256_R) && (in_len == FBC256_256_BLOCK_LEN) && (keylen == FBC256_256_KEY_LEN));

    word* sk = (word *) key;
    word  rk[FBC256_256_SUBKEY_LEN];
    int   rk_num = 2 * cryptround;
    int   i = 0;

    rk[0] = sk[0]; rk[1] = sk[1]; rk[2] = sk[2]; rk[3] = sk[3];
    for (i = 0; i < rk_num - 4; i ++) {
        rk[i+4] = (rk[i] ^ full_one_mask) ^ (((rk[i+1] ^ full_one_mask) & rk[i+2]) & rk[i+3]) ^ i;
        rk[i+4] = L(rk[i+4], 13, 22);
    }

    word  u[4], v[4], x, y;
    word* a = (word *)&output[0];
    word* b = (word *)&input[0];
    a[0] = b[0]; a[1] = b[1]; a[2] = b[2]; a[3] = b[3];
    while (i < rk_num-3) {
        // the left F-function
        x = a[0] ^ rk[i ++];
        F64(x)
        x ^= a[1];

        // the right F-function
        y = a[3] ^ rk[i ++];
        F64(y)
        y ^= a[2];

        a[1] = a[0] ^ y;
        a[2] = a[3] ^ x;
        a[0] = x;
        a[3] = y;
     }

    // the final round
    x = a[0] ^ rk[i ++];
    F64(x)
    a[1] ^= x;

    x = a[3] ^ rk[i];
    F64(x)
    a[2] ^= x;

    a[0] ^= a[2];
    a[3] ^= a[1];

    *out_len = in_len;
    return 0;
}
